#!/usr/bin/env python3
"""Tests for ee.Reducer module."""

import json
from typing import Any, Dict

import unittest
import ee
from ee import apitestcase

TO_LIST = 'Reducer.toList'


def make_expression_graph(
    function_invocation_value: Dict[str, Any],
) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


class ReducerTest(apitestcase.ApiTestCase):

  def test_simple(self):
    reducer = ee.Reducer.toList()
    self.assertEqual({'value': 'fakeValue'}, reducer.getInfo())

    reducer_func = ee.ApiFunction.lookup(TO_LIST)
    self.assertEqual(reducer_func, reducer.func)

    self.assertFalse(reducer.isVariable())
    self.assertEqual({'numOptional': None, 'tupleSize': None}, reducer.args)

    result = json.loads(reducer.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'functionName': TO_LIST,
                    'arguments': {},
                }
            }
        },
    }
    self.assertEqual(expect, result)

    cast_result = json.loads(ee.Reducer(reducer).serialize())
    self.assertEqual(expect, cast_result)

  def test_no_args(self):
    message = r"missing 1 required positional argument: 'reducer'"
    with self.assertRaisesRegex(TypeError, message):
      ee.Reducer()  # pytype: disable=missing-parameter

  def test_bad_arg_literal(self):
    message = (
        r"Reducer can only be used as a cast to Reducer. Found <class 'int'>\."
    )
    with self.assertRaisesRegex(TypeError, message):
      ee.Reducer(1)  # pytype: disable=wrong-arg-types

  def test_all_non_zero(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.allNonZero'}
    )
    expression = ee.Reducer.allNonZero()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_and(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.and'}
    )
    expression = ee.Reducer.And()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_anyNonZero(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.anyNonZero'}
    )
    expression = ee.Reducer.anyNonZero()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_auto_histogram(self):
    max_buckets = 1
    min_bucket_width = 2
    max_raw = 3
    cumulative = True
    expect = make_expression_graph({
        'arguments': {
            'maxBuckets': {'constantValue': max_buckets},
            'minBucketWidth': {'constantValue': min_bucket_width},
            'maxRaw': {'constantValue': max_raw},
            'cumulative': {'constantValue': cumulative},
        },
        'functionName': 'Reducer.autoHistogram',
    })
    expression = ee.Reducer.autoHistogram(
        max_buckets, min_bucket_width, max_raw, cumulative
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.autoHistogram(
        maxBuckets=max_buckets,
        minBucketWidth=min_bucket_width,
        maxRaw=max_raw,
        cumulative=cumulative,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_bitwise_and(self):
    expect = make_expression_graph({
        'arguments': {},
        'functionName': 'Reducer.bitwiseAnd',
    })
    expression = ee.Reducer.bitwiseAnd()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_bitwise_or(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.bitwiseOr'}
    )
    expression = ee.Reducer.bitwiseOr()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_centered_covariance(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.centeredCovariance'}
    )
    expression = ee.Reducer.centeredCovariance()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_circular_mean(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.circularMean'}
    )
    expression = ee.Reducer.circularMean()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_circular_stddev(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.circularStddev'}
    )
    expression = ee.Reducer.circularStddev()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_circular_variance(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.circularVariance'}
    )
    expression = ee.Reducer.circularVariance()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_combine(self):
    reducer1 = ee.Reducer.first()
    reducer2 = ee.Reducer.last()
    output_prefix = 'a'
    shared_inputs = True
    expect = make_expression_graph({
        'arguments': {
            'reducer1': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.first',
                    'arguments': {},
                }
            },
            'reducer2': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
            'outputPrefix': {'constantValue': output_prefix},
            'sharedInputs': {'constantValue': shared_inputs},
        },
        'functionName': 'Reducer.combine',
    })
    expression = reducer1.combine(
        reducer2, output_prefix, shared_inputs
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer1.combine(
        reducer2=reducer2,
        outputPrefix=output_prefix,
        sharedInputs=shared_inputs,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_count(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.count'}
    )
    expression = ee.Reducer.count()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_count_distinct(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.countDistinct'}
    )
    expression = ee.Reducer.countDistinct()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_count_distinct_non_null(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.countDistinctNonNull'}
    )
    expression = ee.Reducer.countDistinctNonNull()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_count_every(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.countEvery'}
    )
    expression = ee.Reducer.countEvery()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_count_runs(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.countRuns'}
    )
    expression = ee.Reducer.countRuns()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_covariance(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.covariance'}
    )
    expression = ee.Reducer.covariance()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_disaggregate(self):
    reducer = ee.Reducer.last()
    axis = 1
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
            'axis': {'constantValue': axis},
        },
        'functionName': 'Reducer.disaggregate',
    })
    expression = reducer.disaggregate(axis)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer.disaggregate(axis=axis)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_first(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.first'}
    )
    expression = ee.Reducer.first()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_first_non_null(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.firstNonNull'}
    )
    expression = ee.Reducer.firstNonNull()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_fixed_2d_histogram(self):
    x_min = 1
    x_max = 2
    x_steps = 3
    y_min = 4
    y_max = 5
    y_steps = 6
    expect = make_expression_graph({
        'arguments': {
            'xMin': {'constantValue': x_min},
            'xMax': {'constantValue': x_max},
            'xSteps': {'constantValue': x_steps},
            'yMin': {'constantValue': y_min},
            'yMax': {'constantValue': y_max},
            'ySteps': {'constantValue': y_steps},
        },
        'functionName': 'Reducer.fixed2DHistogram',
    })
    expression = ee.Reducer.fixed2DHistogram(
        x_min, x_max, x_steps, y_min, y_max, y_steps
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.fixed2DHistogram(
        xMin=x_min,
        xMax=x_max,
        xSteps=x_steps,
        yMin=y_min,
        yMax=y_max,
        ySteps=y_steps,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_fixed_histogram(self):
    min_value = 1
    max_value = 2
    steps = 3
    cumulative = True
    expect = make_expression_graph({
        'arguments': {
            'min': {'constantValue': min_value},
            'max': {'constantValue': max_value},
            'steps': {'constantValue': steps},
            'cumulative': {'constantValue': cumulative},
        },
        'functionName': 'Reducer.fixedHistogram',
    })
    expression = ee.Reducer.fixedHistogram(
        min_value, max_value, steps, cumulative
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.fixedHistogram(
        min=min_value, max=max_value, steps=steps, cumulative=cumulative
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_for_each(self):
    reducer = ee.Reducer.last()
    output_names = ['a', 'b']
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
            'outputNames': {'constantValue': output_names},
        },
        'functionName': 'Reducer.forEach',
    })
    expression = reducer.forEach(output_names)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer.forEach(outputNames=output_names)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_for_each_band(self):
    reducer = ee.Reducer.last()
    image = 'a'
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
            'image': {
                'functionInvocationValue': {
                    'functionName': 'Image.load',
                    'arguments': {'id': {'constantValue': 'a'}},
                }
            },
        },
        'functionName': 'Reducer.forEachBand',
    })
    expression = reducer.forEachBand(image)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer.forEachBand(image=image)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_for_each_element(self):
    reducer = ee.Reducer.last()
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
        },
        'functionName': 'Reducer.forEachElement',
    })
    expression = reducer.forEachElement()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_frequency_histogram(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.frequencyHistogram'}
    )
    expression = ee.Reducer.frequencyHistogram()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_geometric_median(self):
    num_x = 1
    eta = 2
    initial_step_size = 3
    expect = make_expression_graph({
        'arguments': {
            'numX': {'constantValue': num_x},
            'eta': {'constantValue': eta},
            'initialStepSize': {'constantValue': initial_step_size},
        },
        'functionName': 'Reducer.geometricMedian',
    })
    expression = ee.Reducer.geometricMedian(num_x, eta, initial_step_size)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.geometricMedian(
        numX=num_x, eta=eta, initialStepSize=initial_step_size
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_get_outputs(self):
    reducer = ee.Reducer.last()
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
        },
        'functionName': 'Reducer.getOutputs',
    })
    expression = reducer.getOutputs()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_group(self):
    reducer = ee.Reducer.last()
    group_field = 1
    group_name = 'a'
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.last',
                    'arguments': {},
                }
            },
            'groupField': {'constantValue': group_field},
            'groupName': {'constantValue': group_name},
        },
        'functionName': 'Reducer.group',
    })
    expression = reducer.group(group_field, group_name)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer.group(
        groupField=group_field, groupName=group_name
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_histogram(self):
    max_buckets = 1
    min_bucket_width = 2
    max_raw = 3
    expect = make_expression_graph({
        'arguments': {
            'maxBuckets': {'constantValue': max_buckets},
            'minBucketWidth': {'constantValue': min_bucket_width},
            'maxRaw': {'constantValue': max_raw},
        },
        'functionName': 'Reducer.histogram',
    })
    expression = ee.Reducer.histogram(max_buckets, min_bucket_width, max_raw)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.histogram(
        maxBuckets=max_buckets, minBucketWidth=min_bucket_width, maxRaw=max_raw
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_interval_mean(self):
    min_percentile = 1
    max_percentile = 2
    max_buckets = 3
    min_bucket_width = 4
    max_raw = 5
    expect = make_expression_graph({
        'arguments': {
            'minPercentile': {'constantValue': min_percentile},
            'maxPercentile': {'constantValue': max_percentile},
            'maxBuckets': {'constantValue': max_buckets},
            'minBucketWidth': {'constantValue': min_bucket_width},
            'maxRaw': {'constantValue': max_raw},
        },
        'functionName': 'Reducer.intervalMean',
    })
    expression = ee.Reducer.intervalMean(
        min_percentile, max_percentile, max_buckets, min_bucket_width, max_raw
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.intervalMean(
        minPercentile=min_percentile,
        maxPercentile=max_percentile,
        maxBuckets=max_buckets,
        minBucketWidth=min_bucket_width,
        maxRaw=max_raw,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_kendalls_correlation(self):
    num_inputs = 1
    expect = make_expression_graph({
        'arguments': {
            'numInputs': {'constantValue': num_inputs},
        },
        'functionName': 'Reducer.kendallsCorrelation',
    })
    expression = ee.Reducer.kendallsCorrelation(num_inputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.kendallsCorrelation(numInputs=num_inputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_kurtosis(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.kurtosis'}
    )
    expression = ee.Reducer.kurtosis()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_last(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.last'}
    )
    expression = ee.Reducer.last()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_last_non_null(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.lastNonNull'}
    )
    expression = ee.Reducer.lastNonNull()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_linear_fit(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.linearFit'}
    )
    expression = ee.Reducer.linearFit()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_linear_regression(self):
    num_x = 1
    num_y = 2
    expect = make_expression_graph({
        'arguments': {
            'numX': {'constantValue': num_x},
            'numY': {'constantValue': num_y},
        },
        'functionName': 'Reducer.linearRegression',
    })
    expression = ee.Reducer.linearRegression(num_x, num_y)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.linearRegression(numX=num_x, numY=num_y)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_max(self):
    num_inputs = 1
    expect = make_expression_graph({
        'arguments': {
            'numInputs': {'constantValue': num_inputs},
        },
        'functionName': 'Reducer.max',
    })
    expression = ee.Reducer.max(num_inputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.max(numInputs=num_inputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_mean(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.mean'}
    )
    expression = ee.Reducer.mean()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_median(self):
    max_buckets = 1
    min_bucket_width = 2
    max_raw = 3
    expect = make_expression_graph({
        'arguments': {
            'maxBuckets': {'constantValue': max_buckets},
            'minBucketWidth': {'constantValue': min_bucket_width},
            'maxRaw': {'constantValue': max_raw},
        },
        'functionName': 'Reducer.median',
    })
    expression = ee.Reducer.median(max_buckets, min_bucket_width, max_raw)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.median(
        maxBuckets=max_buckets, minBucketWidth=min_bucket_width, maxRaw=max_raw
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_min(self):
    num_inputs = 1
    expect = make_expression_graph({
        'arguments': {
            'numInputs': {'constantValue': num_inputs},
        },
        'functionName': 'Reducer.min',
    })
    expression = ee.Reducer.min(num_inputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.min(numInputs=num_inputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_min_max(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.minMax'}
    )
    expression = ee.Reducer.minMax()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_mode(self):
    max_buckets = 1
    min_bucket_width = 2
    max_raw = 3
    expect = make_expression_graph({
        'arguments': {
            'maxBuckets': {'constantValue': max_buckets},
            'minBucketWidth': {'constantValue': min_bucket_width},
            'maxRaw': {'constantValue': max_raw},
        },
        'functionName': 'Reducer.mode',
    })
    expression = ee.Reducer.mode(max_buckets, min_bucket_width, max_raw)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.mode(
        maxBuckets=max_buckets, minBucketWidth=min_bucket_width, maxRaw=max_raw
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_or(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.or'}
    )
    expression = ee.Reducer.Or()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_pearsons_correlation(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.pearsonsCorrelation'}
    )
    expression = ee.Reducer.pearsonsCorrelation()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_percentile(self):
    percentiles = [1, 2]
    output_names = ['a', 'b']
    max_buckets = 3
    min_bucket_width = 4
    max_raw = 5
    expect = make_expression_graph({
        'arguments': {
            'percentiles': {'constantValue': percentiles},
            'outputNames': {'constantValue': output_names},
            'maxBuckets': {'constantValue': max_buckets},
            'minBucketWidth': {'constantValue': min_bucket_width},
            'maxRaw': {'constantValue': max_raw},
        },
        'functionName': 'Reducer.percentile',
    })
    expression = ee.Reducer.percentile(
        percentiles, output_names, max_buckets, min_bucket_width, max_raw
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.percentile(
        percentiles=percentiles,
        outputNames=output_names,
        maxBuckets=max_buckets,
        minBucketWidth=min_bucket_width,
        maxRaw=max_raw,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_product(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.product'}
    )
    expression = ee.Reducer.product()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_repeat(self):
    reducer = ee.Reducer.product()
    count = 1
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.product',
                    'arguments': {},
                }
            },
            'count': {'constantValue': count},
        },
        'functionName': 'Reducer.repeat',
    })
    expression = reducer.repeat(count)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer.repeat(count=count)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_ridge_regression(self):
    num_x = 1
    num_y = 2
    lambda_ = 3
    expect = make_expression_graph({
        'arguments': {
            'numX': {'constantValue': num_x},
            'numY': {'constantValue': num_y},
            'lambda': {'constantValue': lambda_},
        },
        'functionName': 'Reducer.ridgeRegression',
    })
    expression = ee.Reducer.ridgeRegression(num_x, num_y, lambda_)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    # 'lambda' is a reserved word in Python.
    args = {
        'numX': num_x,
        'numY': num_y,
        'lambda': lambda_,
    }

    expression = ee.Reducer.ridgeRegression(**args)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_robust_linear_regression(self):
    num_x = 1
    num_y = 2
    beta = 3
    expect = make_expression_graph({
        'arguments': {
            'numX': {'constantValue': num_x},
            'numY': {'constantValue': num_y},
            'beta': {'constantValue': beta},
        },
        'functionName': 'Reducer.robustLinearRegression',
    })
    expression = ee.Reducer.robustLinearRegression(num_x, num_y, beta)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.robustLinearRegression(
        numX=num_x, numY=num_y, beta=beta
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_sample_std_dev(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.sampleStdDev'}
    )
    expression = ee.Reducer.sampleStdDev()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_sample_variance(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.sampleVariance'}
    )
    expression = ee.Reducer.sampleVariance()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_sens_slope(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.sensSlope'}
    )
    expression = ee.Reducer.sensSlope()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_set_outputs(self):
    reducer = ee.Reducer.stdDev()
    outputs = ['a', 'b']
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.stdDev',
                    'arguments': {},
                }
            },
            'outputs': {'constantValue': outputs},
        },
        'functionName': 'Reducer.setOutputs',
    })
    expression = reducer.setOutputs(outputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = reducer.setOutputs(outputs=outputs)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_skew(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.skew'}
    )
    expression = ee.Reducer.skew()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_spearmans_correlation(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.spearmansCorrelation'}
    )
    expression = ee.Reducer.spearmansCorrelation()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_split_weights(self):
    reducer = ee.Reducer.stdDev()
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.stdDev',
                    'arguments': {},
                }
            }
        },
        'functionName': 'Reducer.splitWeights',
    })
    expression = reducer.splitWeights()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_std_dev(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.stdDev'}
    )
    expression = ee.Reducer.stdDev()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_sum(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.sum'}
    )
    expression = ee.Reducer.sum()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_to_collection(self):
    property_names = ['a', 'b']
    num_optional = 1
    expect = make_expression_graph({
        'arguments': {
            'propertyNames': {'constantValue': property_names},
            'numOptional': {'constantValue': num_optional},
        },
        'functionName': 'Reducer.toCollection',
    })
    expression = ee.Reducer.toCollection(property_names, num_optional)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.toCollection(
        propertyNames=property_names, numOptional=num_optional
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_to_list(self):
    tuple_size = 1
    num_optional = 2
    expect = make_expression_graph({
        'arguments': {
            'tupleSize': {'constantValue': tuple_size},
            'numOptional': {'constantValue': num_optional},
        },
        'functionName': 'Reducer.toList',
    })
    expression = ee.Reducer.toList(tuple_size, num_optional)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Reducer.toList(
        tupleSize=tuple_size, numOptional=num_optional
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_unweighted(self):
    reducer = ee.Reducer.variance()
    expect = make_expression_graph({
        'arguments': {
            'reducer': {
                'functionInvocationValue': {
                    'functionName': 'Reducer.variance',
                    'arguments': {},
                }
            }
        },
        'functionName': 'Reducer.unweighted',
    })
    expression = reducer.unweighted()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_variance(self):
    expect = make_expression_graph(
        {'arguments': {}, 'functionName': 'Reducer.variance'}
    )
    expression = ee.Reducer.variance()
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



"""A wrapper for Kernels."""
from __future__ import annotations

from typing import Optional

from ee import _arg_types
from ee import apifunction
from ee import computedobject


class Kernel(computedobject.ComputedObject):
  """An object to represent an Earth Engine Kernel.

  Example:
    # Square kernel
    ee.Kernel.square(radius=2, units='pixels', normalize=False)
    # Chebyshev kernel
    ee.Kernel.chebyshev(radius=3)
  """

  _initialized: bool = False

  def __init__(self, kernel: computedobject.ComputedObject):
    """Creates a Kernel wrapper.

    Args:
      kernel: A Kernel to cast.
    """
    self.initialize()

    if isinstance(kernel, computedobject.ComputedObject):
      # There is no server-side constructor for ee.Kernel. Pass the object as-is
      # to the server in case it is intended to be a Kernel cast.
      super().__init__(kernel.func, kernel.args, kernel.varName)
      return

    raise TypeError(
        f'Kernel can only be used as a cast to Kernel. Found {type(kernel)}.'
    )

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Kernel'

  def add(
      self,
      kernel2: _arg_types.Kernel,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns the addition of two kernels.

    Does pointwise addition after aligning their centers.

    Args:
      kernel2: The second kernel.
      normalize: Normalize the kernel.

    Returns:
      An ee.Kernel.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.add', self, kernel2, normalize
    )

  @staticmethod
  def chebyshev(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a Chebyshev kernel.

    Generates a distance kernel based on Chebyshev distance (greatest distance
    along any dimension).

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.chebyshev', radius, units, normalize, magnitude
    )

  @staticmethod
  def circle(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a circle-shaped boolean kernel.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.circle', radius, units, normalize, magnitude
    )

  @staticmethod
  def compass(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 3x3 Prewitt's Compass edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_('Kernel.compass', magnitude, normalize)

  @staticmethod
  def cross(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a cross-shaped boolean kernel.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.cross', radius, units, normalize, magnitude
    )

  @staticmethod
  def diamond(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a diamond-shaped boolean kernel.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.diamond', radius, units, normalize, magnitude
    )

  @staticmethod
  def euclidean(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a distance kernel based on Euclidean (straight-line) distance.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.euclidean', radius, units, normalize, magnitude
    )

  # TODO: Test missing weights.
  @staticmethod
  def fixed(
      width: Optional[_arg_types.Integer] = None,
      height: Optional[_arg_types.Integer] = None,
      weights: Optional[_arg_types.List] = None,
      x: Optional[_arg_types.Integer] = None,
      y: Optional[_arg_types.Integer] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a kernel with the given weights.

    Args:
      width: The width of the kernel in pixels.
      height: The height of the kernel in pixels.
      weights: A 2-D list of [height] x [width] values to use as the weights of
        the kernel.
      x: The location of the focus, as an offset from the left.
      y: The location of the focus, as an offset from the top.
      normalize: Normalize the kernel values to sum to 1.
    """
    if weights is None:
      raise ValueError('weights is required.')

    return apifunction.ApiFunction.call_(
        'Kernel.fixed', width, height, weights, x, y, normalize
    )

  @staticmethod
  def gaussian(
      radius: _arg_types.Number,
      sigma: Optional[_arg_types.Number] = None,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a Gaussian kernel from a sampled continuous Gaussian.

    Args:
      radius: The radius of the kernel to generate.
      sigma: Standard deviation of the Gaussian function (same units as radius).
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.gaussian', radius, sigma, units, normalize, magnitude
    )

  def inverse(self) -> Kernel:
    """Returns a kernel which has each of its weights multiplicatively inverted.

    Weights with a value of zero are not inverted and remain zero.
    """

    return apifunction.ApiFunction.call_(self.name() + '.inverse', self)

  @staticmethod
  def kirsch(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 3x3 Kirsch's Compass edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_('Kernel.kirsch', magnitude, normalize)

  @staticmethod
  def laplacian4(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 3x3 Laplacian-4 edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.laplacian4', magnitude, normalize
    )

  @staticmethod
  def laplacian8(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 3x3 Laplacian-8 edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.laplacian8', magnitude, normalize
    )

  @staticmethod
  def manhattan(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a distance kernel based on rectilinear (city-block) distance.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.manhattan', radius, units, normalize, magnitude
    )

  @staticmethod
  def octagon(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns an octagon-shaped boolean kernel.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.octagon', radius, units, normalize, magnitude
    )

  @staticmethod
  def plus(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a plus-shaped boolean kernel.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.plus', radius, units, normalize, magnitude
    )

  @staticmethod
  def prewitt(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 3x3 Prewitt edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_('Kernel.prewitt', magnitude, normalize)

  @staticmethod
  def rectangle(
      xRadius: _arg_types.Number,  # pylint: disable=invalid-name
      yRadius: _arg_types.Number,  # pylint: disable=invalid-name
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a rectangular-shaped kernel.

    Args:
      xRadius: The horizontal radius of the kernel to generate.
      yRadius: The vertical radius of the kernel to generate.
      units: The system of measurement for the kernel ("pixels" or "meters"). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.rectangle', xRadius, yRadius, units, normalize, magnitude
    )

  @staticmethod
  def roberts(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 2x2 Roberts edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_('Kernel.roberts', magnitude, normalize)

  def rotate(self, rotations: _arg_types.Integer) -> Kernel:
    """Returns a rotated kernel.

    Args:
      rotations: Number of 90 degree rotations to make. Negative numbers rotate
        counterclockwise.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.rotate', self, rotations
    )

  @staticmethod
  def sobel(
      magnitude: Optional[_arg_types.Number] = None,
      normalize: Optional[_arg_types.Bool] = None,
  ) -> Kernel:
    """Returns a 3x3 Sobel edge-detection kernel.

    Args:
      magnitude: Scale each value by this amount.
      normalize: Normalize the kernel values to sum to 1.
    """

    return apifunction.ApiFunction.call_('Kernel.sobel', magnitude, normalize)

  @staticmethod
  def square(
      radius: _arg_types.Number,
      units: Optional[_arg_types.String] = None,
      normalize: Optional[_arg_types.Bool] = None,
      magnitude: Optional[_arg_types.Number] = None,
  ) -> Kernel:
    """Returns a square-shaped boolean kernel.

    Args:
      radius: The radius of the kernel to generate.
      units: The system of measurement for the kernel ('pixels' or 'meters'). If
        the kernel is specified in meters, it will resize when the zoom-level is
        changed.
      normalize: Normalize the kernel values to sum to 1.
      magnitude: Scale each value by this amount.
    """

    return apifunction.ApiFunction.call_(
        'Kernel.square', radius, units, normalize, magnitude
    )



"""A wrapper for dates."""
from __future__ import annotations

import datetime
import math
from typing import Any, Dict, Optional, Union

from ee import _arg_types
from ee import _utils
from ee import apifunction
from ee import computedobject
from ee import daterange
from ee import ee_number
from ee import ee_string
from ee import ee_types as types
from ee import serializer

# _arg_types.DateType does not have datetime.Datetime.
_DateType = Union[
    datetime.datetime, float, str, 'Date', computedobject.ComputedObject
]


class Date(computedobject.ComputedObject):
  """An object to represent dates."""

  _initialized = False

  # Tell pytype to not complain about dynamic attributes.
  _HAS_DYNAMIC_ATTRIBUTES = True

  @_utils.accept_opt_prefix('opt_tz')
  def __init__(
      self, date: Union[_DateType], tz: Optional[_arg_types.String] = None
  ):
    """Construct a date.

    This sends all inputs (except another Date) through the Date function.

    This constructor accepts the following args:
      1) A bare date.
      2) An ISO string
      3) An integer number of milliseconds since the epoch.
      4) A ComputedObject.

    Args:
      date: The date to wrap.
      tz: An optional timezone, only usable with a string date.
    """
    self.initialize()

    func = apifunction.ApiFunction(self.name())
    args: Dict[str, Any]
    var_name = None
    if isinstance(date, datetime.datetime):
      args = {'value':
              math.floor(serializer.DatetimeToMicroseconds(date) / 1000)}
    elif types.isNumber(date):
      args = {'value': date}
    elif isinstance(date, str):
      args = {'value': date}
      if tz:
        if isinstance(tz, (str, computedobject.ComputedObject)):
          args['timeZone'] = tz
        else:
          raise ValueError(
              f'Invalid argument specified for ee.Date(..., opt_tz): {tz}'
          )
    elif isinstance(date, computedobject.ComputedObject):
      if self.is_func_returning_same(date):
        # If it's a call that's already returning a Date, just cast.
        func = date.func
        args = date.args
        var_name = date.varName
      else:
        args = {'value': date}
        if tz:
          if isinstance(tz, (str, computedobject.ComputedObject)):
            args['timeZone'] = tz
          else:
            raise ValueError(
                f'Invalid argument specified for ee.Date(..., opt_tz): {tz}'
            )
    else:
      raise ValueError(f'Invalid argument specified for ee.Date(): {date}')

    super().__init__(func, args, var_name)

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'Date'

  def advance(
      self,
      delta: _arg_types.Number,
      unit: _arg_types.String,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> 'Date':
    """Create a new Date by adding the specified units to the given Date.

    Args:
      delta: The amount to move in the unit. Negative values go back in time.
      unit: One of 'year', 'month', 'week', 'day', 'hour', 'minute', or
        'second'.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.Date.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.advance', self, delta, unit, timeZone
    )

  def difference(
      self, start: _DateType, unit: _arg_types.String
  ) -> ee_number.Number:
    """Returns the difference between two Dates in the specified units.

    Args:
      start: The date to compare to.
      unit: One of 'year', 'month', 'week', 'day', 'hour', 'minute', or
        'second'.

    Returns:
      Returns an ee.Number based on the average length of the unit.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.difference', self, start, unit
    )

  def format(
      self,
      # pylint: disable-next=redefined-builtin
      format: Optional[_arg_types.String] = None,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> ee_string.String:
    """Convert a date to string.

    The format string is described here:

    http://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html

    Args:
      format: A Joda Time pattern. If omitted, it uses the ISO default date.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.String.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.format', self, format, timeZone
    )

  @staticmethod
  def fromYMD(
      year: _arg_types.Integer,
      month: _arg_types.Integer,
      day: _arg_types.Integer,
      # pylint: disable=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> 'Date':
    """Returns a Date given year, month, day.

    Args:
      year: The year, 2013, for example.
      month: The month, 3, for example.
      day: The day, 15, for example.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.
    """

    return apifunction.ApiFunction.call_(
        'Date.fromYMD', year, month, day, timeZone
    )

  def get(
      self,
      unit: _arg_types.String,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> ee_number.Number:
    """Returns the specified unit of this date.

    Args:
      unit: One of 'year', 'month' (returns 1-12), 'week' (1-53), 'day' (1-31),
        'hour' (0-23), 'minute' (0-59), or 'second' (0-59).
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.get', self, unit, timeZone
    )

  def getFraction(
      self,
      unit: _arg_types.String,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> ee_number.Number:
    """Returns this date's elapsed fraction of the specified unit.

    Args:
      unit: One of 'year', 'month', 'week', 'day', 'hour', 'minute', or
        'second'.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.Number between 0 and 1.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.getFraction', self, unit, timeZone
    )

  def getRange(
      self,
      unit: _arg_types.String,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> daterange.DateRange:
    """Returns a DateRange covering the unit that contains this date.

    For example,

      Date('2013-3-15').getRange('year')

    returns

      DateRange('2013-1-1', '2014-1-1').

    Args:
      unit: One of 'year', 'month', 'week', 'day', 'hour', 'minute', or
        'second'.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.DateRange.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.getRange', self, unit, timeZone
    )

  def getRelative(
      self,
      unit: _arg_types.String,
      inUnit: _arg_types.String,  # pylint: disable=invalid-name
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> ee_number.Number:
    """Returns the specified unit of this date relative to a larger unit.

    For example, getRelative('day', 'year') returns a value between 0 and 365.

    Args:
      unit: One of 'month', 'week', 'day', 'hour', 'minute', or 'second'.
      inUnit: One of 'year', 'month', 'week', 'day', 'hour', or 'minute'.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      A 0-based ee.Number.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.getRelative', self, unit, inUnit, timeZone
    )

  def millis(self) -> ee_number.Number:
    """The number of milliseconds since 1970-01-01T00:00:00Z."""

    return apifunction.ApiFunction.call_(self.name() + '.millis', self)

  @staticmethod
  def parse(
      format: _arg_types.String,  # pylint: disable=redefined-builtin
      date: _arg_types.String,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> 'Date':
    """Parse a date string, given a string describing its format.

    Args:
      format: A pattern, as described at
        https://joda-time.sourceforge.net/apidocs/org/joda/time/format/DateTimeFormat.html.
      date: A string matching the given pattern.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.Date.
    """

    return apifunction.ApiFunction.call_('Date.parse', format, date, timeZone)

  @classmethod
  def unitRatio(
      cls, numerator: _arg_types.String, denominator: _arg_types.String
  ) -> ee_number.Number:
    """Returns the ratio of the length of one unit to the length of another.

    For example, unitRatio('day', 'minute') returns 1440.

    Valid units are 'year', 'month', 'week', 'day', 'hour', 'minute', and
    'second'.

    Args:
      numerator: Unit to be divided by the denominator.
      denominator: Unit to divide into the numerator.

    Returns:
      An ee.Number.
    """

    return apifunction.ApiFunction.call_(
        cls.name() + '.unitRatio', numerator, denominator
    )

  def update(
      self,
      year: Optional[_arg_types.Integer] = None,
      month: Optional[_arg_types.Integer] = None,
      day: Optional[_arg_types.Integer] = None,
      hour: Optional[_arg_types.Integer] = None,
      minute: Optional[_arg_types.Integer] = None,
      second: Optional[_arg_types.Number] = None,
      # pylint: disable-next=invalid-name
      timeZone: Optional[_arg_types.String] = None,
  ) -> 'Date':
    """Create a new Date by setting one or more of the units of the given Date.

    If a timeZone is given the new value(s) is interpreted in that zone. Skip or
    set to None any argument to keep the same as the input Date.

    Args:
      year: Set the year.
      month: Set the month.
      day: Set the day.
      hour: Set the hour.
      minute: Set the minute.
      second: Set the second.
      timeZone: The time zone (e.g., 'America/Los_Angeles'); defaults to UTC.

    Returns:
      An ee.Date.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.update',
        self,
        year,
        month,
        day,
        hour,
        minute,
        second,
        timeZone,
    )



#!/usr/bin/env python3
"""Test for the ee.geometry module.

The API objects, like ee.Geometry, are usually tested by comparing the
serialized object to a golden. To be brief, the tests in this file just validate
the Geometry function's arguments instead of the serialized object.
"""

import math

from absl.testing import parameterized

import unittest
import ee
from ee import apitestcase


_EPSILON = 1e-5


class GeometryTest(apitestcase.ApiTestCase, parameterized.TestCase):

  def epsg4326_rectangle(self, west, south, east, north):
    # If we call ee.Geometry.Rectangle with geodesic=False we would get a
    # computed call.
    return ee.Geometry({
        'coordinates': [
            [[west, north], [west, south], [east, south], [east, north]]
        ],
        'type': 'Polygon',
        'geodesic': False,
    })

  @parameterized.named_parameters(
      ('_simple', (-10, -20, 10, 20), (-10, -20, 10, 20)),
      (
          '_latitude_widerThanPolesClamped_south',
          (-10, -90, 10, 73),
          (-10, -1000, 10, 73),
      ),
      (
          '_latitude_widerThanPolesClamped_north',
          (-10, -34, 10, 90),
          (-10, -34, 10, 10000),
      ),
      ('_latitude_zeroSpan', (-10, 20, 10, 20), (-10, 20, 10, 20)),
      (
          '_longitude_crossingMeridianWithOppositeSigns',
          (170, -20, 190, 20),
          (170, -20, -170, 20),
      ),
      (
          '_longitude_crossingMeridianWithNegativeSigns',
          (170, -20, 190, 20),
          (-190, -20, -170, 20),
      ),
      (
          '_longitude_crossingMeridianWithPositiveSigns',
          (170, -20, 190, 20),
          (170, -20, 190, 20),
      ),
      ('_longitude_exactlyGlobal', (-180, -20, 180, 20), (-180, -20, 180, 20)),
      (
          '_longitude_excessOfGlobalIsClamped',
          (-180, -20, 180, 20),
          (-180 - _EPSILON, -20, 180 + _EPSILON, 20),
      ),
      ('_longitude_zeroSpan', (10, -20, 10, 20), (10, -20, 10, 20)),
      (
          '_infinity_validDirection_isClamped_west',
          (-180, -20, 180, 20),
          (-math.inf, -20, 10, 20),
      ),
      (
          '_infinity_validDirection_isClamped_east',
          (-180, -20, 180, 20),
          (-10, -20, math.inf, 20),
      ),
      (
          '_infinity_validDirection_isClamped_south',
          (-10, -90, 10, 20),
          (-10, -math.inf, 10, 20),
      ),
      (
          '_infinity_validDirection_isClamped_north',
          (-10, -20, 10, 90),
          (-10, -20, 10, math.inf),
      ),
  )
  def testBBox(self, expected, coords):
    self.assertEqual(
        self.epsg4326_rectangle(*expected),
        ee.Geometry.BBox(*coords),
    )

  def testBBox_computedArgs(self):
    ten = ee.Number(5).add(5)
    box = ee.Geometry.BBox(-10, -20, ten, 20)
    func = ee.ApiFunction.lookup('GeometryConstructors.BBox')

    self.assertIsInstance(box, ee.Geometry)
    self.assertEqual(func, box.func)
    expected_args = {
        'west': ee.Number(-10),
        'south': ee.Number(-20),
        'east': ten,
        'north': ee.Number(20),
    }
    self.assertEqual(expected_args, box.args)

  @parameterized.named_parameters(
      # Reject cases which, if we clamped them instead, would move a box whose
      # bounds lie past a pole to being a point at the pole.
      (
          '_latitude_notBeyondPoles_north',
          (-10, -100, 10, -95),
          r'Geometry\.BBox: north must be at least -90°, but was -95°',
      ),
      (
          '_latitude_notBeyondPoles_south',
          (-10, 95, 10, 100),
          r'Geometry\.BBox: south must be at most \+90°, but was 95°',
      ),
      # NaN is rejected.
      (
          '_NaN_isRejected_west',
          (math.nan, -20, 10, 20),
          r'Geometry\.BBox: west must not be nan',
      ),
      (
          '_NaN_isRejected_south',
          (-10, math.nan, 10, 20),
          r'Geometry\.BBox: south must be at most \+90°, but was nan°',
      ),
      (
          '_NaN_isRejected_east',
          (-10, -20, math.nan, 20),
          r'Geometry\.BBox: east must not be nan',
      ),
      (
          '_NaN_isRejected_north',
          (-10, -20, 10, math.nan),
          r'Geometry.BBox: north must be at least -90°, but was nan°',
      ),
      # Infinities in the wrong direction are rejected.
      (
          '_infinity_invalidDirection_isRejected_west',
          (math.inf, -20, 10, 20),
          r'Geometry\.BBox: west must not be inf',
      ),
      (
          '_infinity_invalidDirection_isRejected_south',
          (-10, math.inf, 10, 20),
          r'Geometry\.BBox: south must be at most \+90°, but was inf°',
      ),
      (
          '_infinity_invalidDirection_isRejected_east',
          (-10, -20, -math.inf, 20),
          r'Geometry\.BBox: east must not be -inf',
      ),
      (
          '_infinity_invalidDirection_isRejected_north',
          (-10, -20, 10, -math.inf),
          r'Geometry\.BBox: north must be at least -90°, but was -inf°',
      ),
  )
  def testBBoxInvalid(self, coords, message):
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.BBox(*coords)

  def testLineString(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    max_error = 1000
    func = ee.ApiFunction.lookup('GeometryConstructors.LineString')

    geometry = ee.Geometry.LineString(coords, proj, True, max_error)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': True,
            'maxError': ee.ErrorMargin(max_error),
        },
        geometry.args,
    )

  def testLineString_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.LineString')

    geometry = ee.Geometry.LineString(coords, geodesic=True)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': True},
        geometry.args,
    )

  def testLineString_computedArgs(self):
    """Verifies that a LineString with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.LineString([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.LineString'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def testLineStringInvalid(self, coords, message):
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.LineString(*coords)

  def testLinearRing(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    func = ee.ApiFunction.lookup('GeometryConstructors.LinearRing')

    geometry = ee.Geometry.LinearRing(coords, proj, geodesic, max_error)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
        },
        geometry.args,
    )

  def testLinearRing_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.LinearRing')
    geodesic = True

    geometry = ee.Geometry.LinearRing(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def testLinearRing_computedArgs(self):
    """Verifies that a LinearRing with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    ring = ee.Geometry.LinearRing([p1, p2])

    self.assertIsInstance(ring, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.LinearRing'), ring.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, ring.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def testLinearRingInvalid(self, coords, message):
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.LinearRing(*coords)

  def testMultiLineString(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiLineString')

    geometry = ee.Geometry.MultiLineString(coords, proj, geodesic, max_error)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
        },
        geometry.args,
    )

  def testMultiLineString_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiLineString')
    geodesic = True

    geometry = ee.Geometry.MultiLineString(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def testMultiLineString_computedArgs(self):
    """Verifies that a MultiLineString with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.MultiLineString([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.MultiLineString'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[[1, 2], [3, 4]]]]], 'Invalid geometry'),
  )
  def testMultiLineStringInvalid(self, coords, message):
    """Verifies MultiLineString constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.MultiLineString(*coords)

  def testMultiPoint(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPoint')

    geometry = ee.Geometry.MultiPoint(coords, proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
        },
        geometry.args,
    )

  def testMultiPoint_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPoint')
    proj = 'EPSG:4326'

    geometry = ee.Geometry.MultiPoint(coords, proj=proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'crs': ee.Projection(proj)},
        geometry.args,
    )

  def testMultiPoint_computedArgs(self):
    """Verifies that a MultiPoint with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    point = ee.Geometry.MultiPoint([p1, p2])

    self.assertIsInstance(point, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.MultiPoint'), point.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, point.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_nesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def testMultiPointInvalid(self, coords, message):
    """Verifies MultiPoint constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.MultiPoint(*coords)

  def testMultiPolygon(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    even_odd = True
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPolygon')

    geometry = ee.Geometry.MultiPolygon(
        coords, proj, geodesic, max_error, even_odd
    )
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
            'evenOdd': even_odd,
        },
        geometry.args,
    )

  def testMultiPolygon_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.MultiPolygon')
    geodesic = True

    geometry = ee.Geometry.MultiPolygon(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def testMultiPolygon_computedArgs(self):
    """Verifies that a MultiPolygon with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.MultiPolygon([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.MultiPolygon'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[[[[1, 2], [3, 4], [5, 6]]]]]], 'Invalid geometry'),
      ('_nesting', [[[[[1, 2], [3, 4]], [1, 2]]]], 'Invalid geometry'),
  )
  def testMultiPolygonInvalid(self, coords, message):
    """Verifies MultiPolygon constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.MultiPolygon(*coords)

  def testPoint(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    func = ee.ApiFunction.lookup('GeometryConstructors.Point')

    geometry = ee.Geometry.Point(coords, proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
        },
        geometry.args,
    )

  def testPoint_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.Point')
    proj = 'EPSG:4326'

    geometry = ee.Geometry.Point(coords, proj=proj)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'crs': ee.Projection(proj)},
        geometry.args,
    )

  def testPoint_oldKeywordArgs(self):
    """Verifies that Points still allow keyword lon/lat args."""
    self.assertEqual(ee.Geometry.Point(1, 2), ee.Geometry.Point(lon=1, lat=2))
    self.assertEqual(ee.Geometry.Point(1, 2), ee.Geometry.Point(1, lat=2))

  def testPoint_computedArgs(self):
    """Verifies that a Point with computed inputs can be constructed."""
    v1 = ee.Number(1)
    v2 = ee.Number(2)
    point = ee.Geometry.Point([v1, v2])

    self.assertIsInstance(point, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), point.func
    )
    self.assertEqual({'coordinates': ee.List([v1, v2])}, point.args)

  @parameterized.named_parameters(
      ('_geometry', [['-78.204948', '40.966539']], 'Invalid geometry'),
  )
  def testPointInvalid(self, coords, message):
    """Verifies Point constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.Point(*coords)

  def testPolygon(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    max_error = 1000
    even_odd = True
    func = ee.ApiFunction.lookup('GeometryConstructors.Polygon')

    geometry = ee.Geometry.Polygon(coords, proj, geodesic, max_error, even_odd)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'maxError': ee.ErrorMargin(max_error),
            'evenOdd': even_odd,
        },
        geometry.args,
    )

  def testPolygon_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.Polygon')
    geodesic = True

    geometry = ee.Geometry.Polygon(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def testPolygon_computedArgs(self):
    """Verifies that a Polygon with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.Polygon([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Polygon'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  def testPolygon_evenOdd(self):
    poly1 = ee.Geometry.Polygon([0, 0, 0, 5, 5, 0])
    self.assertTrue(poly1.toGeoJSON()['evenOdd'])

    # TODO: Use kwargs instead of positional args for evenOdd.
    poly2 = ee.Geometry.Polygon([0, 0, 0, 5, 5, 0], None, None, None, False)
    self.assertFalse(poly2.toGeoJSON()['evenOdd'])

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_tooMuchNesting', [[[[[1, 2], [3, 4], [5, 6]]]]], 'Invalid geometry'),
      ('_badNesting', [[[[1, 2], [3, 4]], [1, 2]]], 'Invalid geometry'),
  )
  def testPolygonInvalid(self, coords, message):
    """Verifies Polygon constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.Polygon(*coords)

  def testRectangle(self):
    coords = [1, 2, 3, 4]
    proj = 'EPSG:4326'
    geodesic = True
    even_odd = True
    func = ee.ApiFunction.lookup('GeometryConstructors.Rectangle')

    geometry = ee.Geometry.Rectangle(coords, proj, geodesic, even_odd)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {
            'coordinates': ee.List(coords),
            'crs': ee.Projection(proj),
            'geodesic': geodesic,
            'evenOdd': even_odd,
        },
        geometry.args,
    )

  def testRectangle_kwargs(self):
    coords = [1, 2, 3, 4]
    func = ee.ApiFunction.lookup('GeometryConstructors.Rectangle')
    geodesic = True

    geometry = ee.Geometry.Rectangle(coords, geodesic=geodesic)
    self.assertEqual(func, geometry.func)
    self.assertEqual(
        {'coordinates': ee.List(coords), 'geodesic': geodesic},
        geometry.args,
    )

  def testRectangle_oldKeywordArgs(self):
    """Verifies that Rectangles still allow keyword xlo/ylo/xhi/yhi args."""
    self.assertEqual(
        ee.Geometry.Rectangle(1, 2, 3, 4),
        ee.Geometry.Rectangle(xlo=1, ylo=2, xhi=3, yhi=4),
    )
    self.assertEqual(
        ee.Geometry.Rectangle(1, 2, 3, 4),
        ee.Geometry.Rectangle(1, 2, xhi=3, yhi=4),
    )

  def testRectangle_computedArgs(self):
    """Verifies that a Rectangle with computed inputs can be constructed."""
    p1 = ee.Geometry.Point([1, 2])
    p2 = ee.Geometry.Point([3, 4])
    line = ee.Geometry.Rectangle([p1, p2])

    self.assertIsInstance(line, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Rectangle'), line.func
    )
    self.assertEqual({'coordinates': ee.List([p1, p2])}, line.args)

  @parameterized.named_parameters(
      ('_coords', (1, 2, 3, 4, 5), 'Invalid number of coordinates: 5'),
      ('_coordsList', [[1, 2, 3, 4, 5]], 'Invalid number of coordinates: 5'),
      ('_geometry', [[[1, 2], [3, 4], 5]], 'Invalid geometry'),
      ('_tooMuchNesting', [[[[1, 2], [3, 4]]]], 'Invalid geometry'),
  )
  def testRectangleInvalid(self, coords, message):
    """Verifies Rectangle constructor behavior with invalid arguments."""
    with self.assertRaisesRegex(ee.EEException, message):
      ee.Geometry.Rectangle(*coords)

  def testValid_GeometryCollection(self):
    """Verifies GeometryCollection constructor behavior with valid arguments."""
    geometry = ee.Geometry({
        'type': 'GeometryCollection',
        'geometries': [
            {
                'type': 'Polygon',
                'coordinates': [[[-1, -1], [0, 1], [1, -1]]],
                'geodesic': True,
                'evenOdd': True,
            },
            {'type': 'Point', 'coordinates': [0, 0]},
            {
                'type': 'GeometryCollection',
                'geometries': [
                    {'type': 'Point', 'coordinates': [1, 2]},
                    {'type': 'Point', 'coordinates': [2, 1]},
                ],
            },
        ],
        'coordinates': [],
    })
    self.assertIsInstance(geometry, ee.Geometry)

  def testArrayConstructors(self):
    """Verifies that constructors that take arrays fix nesting."""
    get_coordinates_count = lambda g: len(g.toGeoJSON()['coordinates'])

    point = ee.Geometry.Point([1, 2])
    self.assertEqual(2, get_coordinates_count(point))

    multipoint = ee.Geometry.MultiPoint([[1, 2], [3, 4], [5, 6]])
    self.assertEqual(3, get_coordinates_count(multipoint))

    line = ee.Geometry.LineString([[1, 2], [3, 4], [5, 6]])
    self.assertEqual(3, get_coordinates_count(line))

    ring = ee.Geometry.LinearRing([[1, 2], [3, 4], [5, 6]])
    self.assertEqual(3, get_coordinates_count(ring))

    multiline = ee.Geometry.MultiLineString(
        [[[1, 2], [3, 4]], [[5, 6], [7, 8]]]
    )
    self.assertEqual(2, get_coordinates_count(multiline))

    polygon = ee.Geometry.Polygon([[[1, 2], [3, 4], [5, 6]]])
    self.assertEqual(1, get_coordinates_count(polygon))

    mpolygon = ee.Geometry.MultiPolygon(
        [[[[1, 2], [3, 4], [5, 6]]], [[[1, 2], [3, 4], [5, 6]]]]
    )
    self.assertEqual(2, get_coordinates_count(mpolygon))

  def testGeodesicFlag(self):
    """Verifies that JSON parsing and generation preserves the geodesic flag."""
    geodesic = ee.Geometry({
        'type': 'LineString',
        'coordinates': [[1, 2], [3, 4]],
        'geodesic': True,
    })
    projected = ee.Geometry({
        'type': 'LineString',
        'coordinates': [[1, 2], [3, 4]],
        'geodesic': False,
    })
    self.assertTrue(geodesic.toGeoJSON()['geodesic'])
    self.assertFalse(projected.toGeoJSON()['geodesic'])

  def testConstructor(self):
    """Check the behavior of the Geometry constructor.

    There are 5 options:
      1) A geoJSON object.
      2) A not-computed geometry.
      3) A not-computed geometry with overrides.
      4) A computed geometry.
      5) something to cast to geometry.
    """
    line = ee.Geometry.LineString(1, 2, 3, 4)

    # GeoJSON.
    from_json = ee.Geometry(line.toGeoJSON())
    self.assertIsNone(from_json.func)
    self.assertEqual(from_json._type, 'LineString')
    self.assertEqual(from_json._coordinates, [[1, 2], [3, 4]])

    # GeoJSON with a CRS specified.
    json_with_crs = line.toGeoJSON()
    json_with_crs['crs'] = {
        'type': 'name',
        'properties': {'name': 'SR-ORG:6974'},
    }
    from_json_with_crs = ee.Geometry(json_with_crs)
    self.assertIsNone(from_json_with_crs.func)
    self.assertEqual(from_json_with_crs._type, 'LineString')
    self.assertEqual(from_json_with_crs._proj, 'SR-ORG:6974')

    # A not-computed geometry.
    self.assertEqual(ee.Geometry(line), line)

    # A not-computed geometry with an override.
    with_override = ee.Geometry(line, 'SR-ORG:6974')
    self.assertEqual(with_override._proj, 'SR-ORG:6974')

    # A computed geometry.
    self.assertEqual(ee.Geometry(line.bounds()), line.bounds())

    # Something to cast to a geometry.
    computed = ee.ComputedObject(ee.Function(), {'a': 1})
    geom = ee.Geometry(computed)
    self.assertEqual(computed.func, geom.func)
    self.assertEqual(computed.args, geom.args)

  def testComputedGeometries(self):
    """Verifies the computed object behavior of the Geometry constructor."""
    line = ee.Geometry.LineString(1, 2, 3, 4)
    bounds = line.bounds()

    self.assertIsInstance(bounds, ee.Geometry)
    self.assertEqual(ee.ApiFunction.lookup('Geometry.bounds'), bounds.func)
    self.assertEqual(line, bounds.args['geometry'])
    self.assertTrue(hasattr(bounds, 'bounds'))

  def testComputedCoordinate(self):
    """Verifies that a computed coordinate produces a computed geometry."""
    coords = [1, ee.Number(1).add(1)]
    p = ee.Geometry.Point(coords)

    self.assertIsInstance(p, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), p.func
    )
    self.assertEqual({'coordinates': ee.List(coords)}, p.args)

  def testComputedList(self):
    """Verifies that a computed coordinate produces a computed geometry."""
    lst = ee.List([1, 2, 3, 4]).slice(0, 2)
    p = ee.Geometry.Point(lst)

    self.assertIsInstance(p, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), p.func
    )
    self.assertEqual({'coordinates': lst}, p.args)

  def testComputedProjection(self):
    """Verifies that a geometry with a projection can be constructed."""
    p = ee.Geometry.Point([1, 2], 'epsg:4326')

    self.assertIsInstance(p, ee.Geometry)
    self.assertEqual(
        ee.ApiFunction.lookup('GeometryConstructors.Point'), p.func
    )
    expected_args = {
        'coordinates': ee.List([1, 2]),
        'crs': ee.ApiFunction.lookup('Projection').call('epsg:4326'),
    }
    self.assertEqual(expected_args, p.args)

  def testInvalidCrs(self):
    geo_json = ee.Geometry.LineString(1, 2, 3, 4).toGeoJSON()
    geo_json['crs'] = {'something': 'invalid-crs'}
    with self.assertRaisesRegex(ee.EEException, 'Invalid CRS declaration'):
      ee.Geometry(geo_json)

  def testInternals(self):
    """Test eq(), ne() and hash()."""
    a = ee.Geometry.Point(1, 2)
    b = ee.Geometry.Point(2, 1)
    c = ee.Geometry.Point(1, 2)

    self.assertEqual(a, a)
    self.assertNotEqual(a, b)
    self.assertEqual(a, c)
    self.assertNotEqual(b, c)
    self.assertNotEqual(hash(a), hash(b))

  def testInit_optParams(self):
    result = ee.Geometry(
        geo_json={'type': 'Polygon', 'coordinates': [[[-2, 1]]]},
        opt_proj='abc',
        opt_geodesic=True,
        opt_evenOdd=True,
    ).serialize()
    self.assertIn('"crs": {"constantValue": "abc"}', result)
    self.assertIn('"geodesic": {"constantValue": true}', result)
    self.assertIn('"evenOdd": {"constantValue": true}', result)

  @parameterized.named_parameters(
      (
          'BBox',
          lambda: ee.Geometry.BBox(1, 2, 3, 4),
          dict(
              coordinates=[[[1, 4], [1, 2], [3, 2], [3, 4]]],
              geodesic=False,
              type='Polygon',
          ),
      ),
      (
          'LineString',
          lambda: ee.Geometry.LineString(1, 2, 3, 4),
          dict(coordinates=[[1, 2], [3, 4]], type='LineString'),
      ),
      (
          'LinearRing',
          lambda: ee.Geometry.LinearRing(1, 2, 3, 4),
          dict(coordinates=[[1, 2], [3, 4]], type='LinearRing'),
      ),
      (
          'MultiLineString',
          lambda: ee.Geometry.MultiLineString(1, 2, 3, 4),
          dict(coordinates=[[[1, 2], [3, 4]]], type='MultiLineString'),
      ),
      (
          'MultiLineString_Empty',
          lambda: ee.Geometry.MultiLineString(),  # pylint: disable=unnecessary-lambda
          dict(coordinates=[], type='MultiLineString'),
      ),
      (
          'MultiPoint',
          lambda: ee.Geometry.MultiPoint(1, 2, 3, 4),
          dict(coordinates=[[1, 2], [3, 4]], type='MultiPoint'),
      ),
      (
          'MultiPoint_Empty',
          lambda: ee.Geometry.MultiPoint(),  # pylint: disable=unnecessary-lambda
          dict(coordinates=[], type='MultiPoint'),
      ),
      (
          'MultiPolygon',
          lambda: ee.Geometry.MultiPolygon(1, 2, 3, 4),
          dict(
              coordinates=[[[[1, 2], [3, 4]]]],
              evenOdd=True,
              type='MultiPolygon',
          ),
      ),
      (
          'MultiPolygon_Empty',
          lambda: ee.Geometry.MultiPolygon(),  # pylint: disable=unnecessary-lambda
          dict(coordinates=[], evenOdd=True, type='MultiPolygon'),
      ),
      (
          'Point',
          lambda: ee.Geometry.Point(coords=[1, 2]),
          dict(coordinates=[1, 2], type='Point'),
      ),
      (
          'Polygon',
          lambda: ee.Geometry.Polygon(1, 2, 3, 4),
          dict(
              coordinates=[[[1, 2], [3, 4]]],
              evenOdd=True,
              type='Polygon',
          ),
      ),
      (
          'Rectangle',
          lambda: ee.Geometry.Rectangle(1, 2, 3, 4),
          dict(
              coordinates=[[[1, 4], [1, 2], [3, 2], [3, 4]]],
              evenOdd=True,
              type='Polygon',
          ),
      ),
  )
  def testToGeoJSON(self, geometry_func, expected_geojson):
    self.assertEqual(expected_geojson, geometry_func().toGeoJSON())

  @parameterized.named_parameters(
      # BBox doesn't have any other arguments.
      ('LineString', lambda: ee.Geometry.LineString([], 'EPSG:4326')),
      ('LinearRing', lambda: ee.Geometry.LinearRing([], 'EPSG:4326')),
      ('MultiLineString', lambda: ee.Geometry.MultiLineString([], 'EPSG:4326')),
      ('MultiPoint', lambda: ee.Geometry.MultiPoint([], 'EPSG:4326')),
      ('MultiPolygon', lambda: ee.Geometry.MultiPolygon([], 'EPSG:4326')),
      ('Point', lambda: ee.Geometry.Point([], 'EPSG:4326')),
      ('Polygon', lambda: ee.Geometry.Polygon([], 'EPSG:4326')),
      ('Rectangle', lambda: ee.Geometry.Rectangle([], 'EPSG:4326')),
  )
  def testToGeoJSON_FailsOnComputed(self, geometry_func):
    message = 'Cannot convert a computed geometry to GeoJSON'
    with self.assertRaisesRegex(ee.EEException, message):
      geometry_func().toGeoJSON()


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Tests for the ee.Model module."""

import json
from typing import Any, Dict

import unittest
import ee
from ee import apitestcase

MODEL = 'Model'


def make_expression_graph(
    function_invocation_value: Dict[str, Any],
) -> Dict[str, Any]:
  return {
      'result': '0',
      'values': {'0': {'functionInvocationValue': function_invocation_value}},
  }


def make_override_expression(key: str, pixel_type: str) -> Dict[str, Any]:
  return {
      'dictionaryValue': {
          'values': {
              key: {
                  'functionInvocationValue': {
                      'functionName': 'PixelType',
                      'arguments': {
                          'precision': {
                              'functionInvocationValue': {
                                  'functionName': pixel_type,
                                  'arguments': {},
                              }
                          }
                      },
                  }
              }
          }
      }
  }


def make_type_expression(
    key: str, pixel_type: str, dimensions: int
) -> Dict[str, Any]:
  return {
      'dictionaryValue': {
          'values': {
              key: {
                  'dictionaryValue': {
                      'values': {
                          'dimensions': {'constantValue': dimensions},
                          'type': {
                              'functionInvocationValue': {
                                  'functionName': 'PixelType',
                                  'arguments': {
                                      'precision': {
                                          'functionInvocationValue': {
                                              'functionName': pixel_type,
                                              'arguments': {},
                                          }
                                      }
                                  },
                              }
                          },
                      }
                  }
              }
          }
      }
  }


class ModelTest(apitestcase.ApiTestCase):

  def test_serialize(self):
    model = ee.Model.fromAiPlatformPredictor('some-project', 'some-model')
    result = json.loads(model.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'functionName': 'Model.fromAiPlatformPredictor',
                    'arguments': {
                        'projectId': {'constantValue': 'some-model'},
                        'projectName': {'constantValue': 'some-project'},
                    },
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_cast(self):
    model = ee.Model(
        ee.Model.fromAiPlatformPredictor('some-project', 'some-model')
    )
    result = json.loads(model.serialize())
    expect = {
        'result': '0',
        'values': {
            '0': {
                'functionInvocationValue': {
                    'functionName': 'Model.fromAiPlatformPredictor',
                    'arguments': {
                        'projectId': {'constantValue': 'some-model'},
                        'projectName': {'constantValue': 'some-project'},
                    },
                }
            }
        },
    }
    self.assertEqual(expect, result)

  def test_from_ai_platform_predictor(self):
    project_name = 'some project'
    project_id = 'a project id'
    model_name = 'model name'
    version = 'a version'
    region = 'us-central1'
    input_properties = ['a', 'b']
    input_type_override = {'c': ee.PixelType.int8()}
    input_shapes = {'d': [1, 2]}
    proj = 'epsg:4326'
    fix_input_proj = True
    input_tile_size = [3, 4]
    input_overlap_size = [5, 6]
    output_tile_size = [7, 8]
    output_bands = {'e': {'type': ee.PixelType.int16(), 'dimensions': 10}}
    output_properties = {'f': {'type': ee.PixelType.int32(), 'dimensions': 11}}
    output_multiplier = 9.1

    expect = make_expression_graph({
        'functionName': 'Model.fromAiPlatformPredictor',
        'arguments': {
            'fixInputProj': {'constantValue': fix_input_proj},
            'inputOverlapSize': {'constantValue': input_overlap_size},
            'inputProperties': {'constantValue': input_properties},
            'inputShapes': {'constantValue': input_shapes},
            'inputTileSize': {'constantValue': input_tile_size},
            'inputTypeOverride': make_override_expression(
                'c', 'PixelType.int8'
            ),
            'modelName': {'constantValue': model_name},
            'outputBands': make_type_expression('e', 'PixelType.int16', 10),
            'outputMultiplier': {'constantValue': output_multiplier},
            'outputProperties': make_type_expression(
                'f', 'PixelType.int32', 11
            ),
            'outputTileSize': {'constantValue': output_tile_size},
            'proj': {
                'functionInvocationValue': {
                    'functionName': 'Projection',
                    'arguments': {'crs': {'constantValue': 'epsg:4326'}},
                }
            },
            'projectId': {'constantValue': project_id},
            'projectName': {'constantValue': project_name},
            'region': {'constantValue': region},
            'version': {'constantValue': version},
        },
    })
    expression = ee.Model.fromAiPlatformPredictor(
        project_name,
        project_id,
        model_name,
        version,
        region,
        input_properties,
        input_type_override,
        input_shapes,
        proj,
        fix_input_proj,
        input_tile_size,
        input_overlap_size,
        output_tile_size,
        output_bands,
        output_properties,
        output_multiplier,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Model.fromAiPlatformPredictor(
        projectName=project_name,
        projectId=project_id,
        modelName=model_name,
        version=version,
        region=region,
        inputProperties=input_properties,
        inputTypeOverride=input_type_override,
        inputShapes=input_shapes,
        proj=proj,
        fixInputProj=fix_input_proj,
        inputTileSize=input_tile_size,
        inputOverlapSize=input_overlap_size,
        outputTileSize=output_tile_size,
        outputBands=output_bands,
        outputProperties=output_properties,
        outputMultiplier=output_multiplier,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_from_vertex_ai(self):
    project = 'ee-demos'
    region = 'us-central1'
    endpoint_id = '5184137951335940096'
    endpoint = f'projects/{project}/locations/{region}/endpoints/{endpoint_id}'

    input_properties = ['a', 'b']
    input_type_override = {'c': ee.PixelType.int8()}
    input_shapes = {'d': [1, 2]}
    proj = 'epsg:4326'
    fix_input_proj = True
    input_tile_size = [3, 4]
    input_overlap_size = [5, 6]
    output_tile_size = [7, 8]
    output_bands = {'e': {'type': ee.PixelType.int16(), 'dimensions': 10}}
    output_properties = {'f': {'type': ee.PixelType.int32(), 'dimensions': 11}}
    output_multiplier = 9.1
    max_payload_bytes = 2014
    payload_format = 'RAW_JSON'

    expect = make_expression_graph({
        'functionName': 'Model.fromVertexAi',
        'arguments': {
            'endpoint': {'constantValue': endpoint},
            'fixInputProj': {'constantValue': fix_input_proj},
            'inputOverlapSize': {'constantValue': input_overlap_size},
            'inputProperties': {'constantValue': input_properties},
            'inputShapes': {'constantValue': input_shapes},
            'inputTileSize': {'constantValue': input_tile_size},
            'inputTypeOverride': make_override_expression(
                'c', 'PixelType.int8'
            ),
            'maxPayloadBytes': {'constantValue': max_payload_bytes},
            'outputBands': make_type_expression('e', 'PixelType.int16', 10),
            'outputMultiplier': {'constantValue': output_multiplier},
            'outputProperties': make_type_expression(
                'f', 'PixelType.int32', 11
            ),
            'outputTileSize': {'constantValue': output_tile_size},
            'payloadFormat': {'constantValue': payload_format},
            'proj': {
                'functionInvocationValue': {
                    'functionName': 'Projection',
                    'arguments': {'crs': {'constantValue': 'epsg:4326'}},
                }
            },
        },
    })
    expression = ee.Model.fromVertexAi(
        endpoint,
        input_properties,
        input_type_override,
        input_shapes,
        proj,
        fix_input_proj,
        input_tile_size,
        input_overlap_size,
        output_tile_size,
        output_bands,
        output_properties,
        output_multiplier,
        max_payload_bytes,
        payload_format,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = ee.Model.fromVertexAi(
        endpoint=endpoint,
        inputProperties=input_properties,
        inputTypeOverride=input_type_override,
        inputShapes=input_shapes,
        proj=proj,
        fixInputProj=fix_input_proj,
        inputTileSize=input_tile_size,
        inputOverlapSize=input_overlap_size,
        outputTileSize=output_tile_size,
        outputBands=output_bands,
        outputProperties=output_properties,
        outputMultiplier=output_multiplier,
        maxPayloadBytes=max_payload_bytes,
        payloadFormat=payload_format,
    )
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_predict_image(self):
    end_point = 'an end point'
    model = ee.Model.fromVertexAi(end_point)
    image_name = 'an image'
    image = ee.Image(image_name)

    expect = make_expression_graph({
        'arguments': {
            'model': {
                'functionInvocationValue': {
                    'functionName': 'Model.fromVertexAi',
                    'arguments': {'endpoint': {'constantValue': end_point}},
                }
            },
            'image': {
                'functionInvocationValue': {
                    'functionName': 'Image.load',
                    'arguments': {'id': {'constantValue': image_name}},
                }
            },
        },
        'functionName': 'Model.predictImage',
    })
    expression = model.predictImage(image)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = model.predictImage(image=image)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

  def test_predict_properties(self):
    end_point = 'an end point'
    model = ee.Model.fromVertexAi(end_point)
    feature_collection_name = 'a feature collection'
    collection = fc = ee.FeatureCollection(feature_collection_name)
    expect = make_expression_graph({
        'arguments': {
            'model': {
                'functionInvocationValue': {
                    'functionName': 'Model.fromVertexAi',
                    'arguments': {'endpoint': {'constantValue': end_point}},
                }
            },
            'collection': {
                'functionInvocationValue': {
                    'functionName': 'Collection.loadTable',
                    'arguments': {
                        'tableId': {'constantValue': 'a feature collection'}
                    },
                }
            },
        },
        'functionName': 'Model.predictProperties',
    })
    expression = model.predictProperties(collection)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)

    expression = model.predictProperties(collection=collection)
    result = json.loads(expression.serialize())
    self.assertEqual(expect, result)


if __name__ == '__main__':
  unittest.main()



"""A wrapper for ConfusionMatrices."""
from __future__ import annotations

from typing import Any, Dict, Optional, Union

from ee import _arg_types
from ee import apifunction
from ee import computedobject
from ee import ee_array
from ee import ee_list
from ee import ee_number


class ConfusionMatrix(computedobject.ComputedObject):
  """An object to represent an Earth Engine ConfusionMatrix.

  Examples:
    # Construct a confusion matrix. Rows correspond to actual values, columns to
    # predicted values.
    matrix = ee.Array([[32, 0, 0,  0,  1, 0],
                       [ 0, 5, 0,  0,  1, 0],
                       [ 0, 0, 1,  3,  0, 0],
                       [ 0, 1, 4, 26,  8, 0],
                       [ 0, 0, 0,  7, 15, 0],
                       [ 0, 0, 0,  1,  0, 5]])
    ee.ConfusionMatrix(matrix)
  """

  _initialized: bool = False

  def __init__(
      self,
      array: Union[
          ee_array.Array,
          _arg_types.ConfusionMatrix,
          computedobject.ComputedObject,
      ],
      order: Optional[_arg_types.List] = None,
  ):
    """Creates a ConfusionMatrix wrapper.

    Creates a confusion matrix. Axis 0 (the rows) of the matrix correspond to
    the actual values, and Axis 1 (the columns) to the predicted values.

    Args:
      array: A square, 2D array of integers, representing the confusion matrix.
        Note that unlike the ee.Array constructor, this argument cannot take a
        list.
      order: The row and column size and order, for non-contiguous or non-zero
        based matrices.
    """
    self.initialize()

    if isinstance(array, computedobject.ComputedObject) and order is None:
      if self.is_func_returning_same(array):
        # If it is a call that already returns a ConfusionMatrix, just cast.
        super().__init__(array.func, array.args, array.varName)
        return

    args: Dict[str, Any] = {'array': array}
    if order is not None:
      args['order'] = order

    func = apifunction.ApiFunction(self.name())
    super().__init__(func, func.promoteArgs(args))

  @classmethod
  def initialize(cls) -> None:
    """Imports API functions to this class."""
    if not cls._initialized:
      apifunction.ApiFunction.importApi(cls, cls.name(), cls.name())
      cls._initialized = True

  @classmethod
  def reset(cls) -> None:
    """Removes imported API functions from this class."""
    apifunction.ApiFunction.clearApi(cls)
    cls._initialized = False

  @staticmethod
  def name() -> str:
    return 'ConfusionMatrix'

  def accuracy(self) -> ee_number.Number:
    """Returns the accuracy of a confusion matrix defined as correct / total."""

    return apifunction.ApiFunction.call_(self.name() + '.accuracy', self)

  def array(self) -> ee_array.Array:
    """Returns a confusion matrix as an Array."""

    return apifunction.ApiFunction.call_(self.name() + '.array', self)

  def consumersAccuracy(self) -> ee_array.Array:
    """Returns an array of consumer's accuracies.

    Computes the consumer's accuracy (reliability) of a confusion matrix defined
    as (correct / total) for each row.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.consumersAccuracy', self
    )

  def fscore(self, beta: Optional[_arg_types.Number] = None) -> ee_array.Array:
    """Returns the F-beta score for the confusion matrix.

    Args:
      beta: A factor indicating how much more important recall is than
        precision. The standard F-score is equivalent to setting beta to one.
    """

    return apifunction.ApiFunction.call_(self.name() + '.fscore', self, beta)

  def kappa(self) -> ee_number.Number:
    """Returns the Kappa statistic for the confusion matrix."""

    return apifunction.ApiFunction.call_(self.name() + '.kappa', self)

  def order(self) -> ee_list.List:
    """Returns the name and order of the rows and columns of the matrix."""

    return apifunction.ApiFunction.call_(self.name() + '.order', self)

  def producersAccuracy(self) -> ee_array.Array:
    """Returns an array of producer's accuracies.

    Computes the producer's accuracy of a confusion matrix defined as (correct /
    total) for each column.
    """

    return apifunction.ApiFunction.call_(
        self.name() + '.producersAccuracy', self
    )
